// scanner.bas
// Lexical analyer and symbol table for BASICO
//
// The symbol tables are 1-dim char arrays. The symbols
// are space terminated.
// symbol tables are
// keywords for reserved words
// globals for global vars
// locals for for formal parameters and local vars
//
// Search first keywords, then locals, then globals
//
// by Andre Adrian
//
// version 20jun2006
//
// ./basico05<scanner.bas >scanner.s
// cc -o scanner scanner.s
 
 
 
 
 
 
// scanner
var Look: int                    // Lookahead Character
  inputl: array 128 char         // Input line
  inputi: int  
  yytext: array 128 char         // identifier or string or number, ..
  token:  array 128 char         // identifier or string
  number: int                    // int const
  nows:   int                    // No White Space flag
  
// symtab  
  keywords:     array 100 char   
  tokens:       array 19 char 
  globals:      array 1000 char 
  globaltype:   array 100 char 
  locals:       array 1000 char 
  localtype:    array 100 char 
  localoffset:  array 100 int  
 
func error(msg: array char): void
begin
  inputl[inputi] = 0
  printf("\n# %s\n", inputl)
  printf("# error %s\n", msg)
  exit(1)
end
  
// Read New Character From Input Stream (see Crenshaw)
func GetChar(): void
begin
  if Look = '\n' then
    inputl[inputi] = 0
    printf("# %s", inputl)
    inputi = 0
  endif  
  Look = getchar()
  if Look >= 0 then
    inputl[inputi] = Look
    inputi = inputi + 1
  endif
end

// Initialize (see Crenshaw)
func scannerInit(): void
begin
  inputi = 0
  GetChar()
end

// return 1 if ch is in s, 0 else
func in(ch: char, s: array char): int
var i: int
begin
  i = 0
  while s[i] # 0 do
    if ch = s[i] then
      return 1
    endif
    i = i + 1
  wend
  return 0
end

// scanner function
func yylex(): int
var i: int
  ch: char
begin
  yytext[0] = 0
  
  // linefeed to ; translation is part of skip white space
  
  // ignore white space
  while (Look = ' ') | (Look = '\t') | (Look = '\n') do
    if (Look = '\n') & nows then
      nows = 0
      return ';'
    endif
    GetChar()
  wend
  
  // ignore comment
  while Look = '/' do
    GetChar()
    if Look # '/' then
      return '/'
    endif    
    while Look # '\n' do
      GetChar()   
    wend
    if nows then
      nows = 0
      return ';'
    endif
    GetChar()
    while (Look = ' ') | (Look = '\t') | (Look = '\n') do
      GetChar()
    wend
  wend
    
  // one character tokens
  if in(Look, "-+*/%&|^~,:;()=#<>[]") then
    nows = 1
    ch = Look
    GetChar()
    return ch
  endif
  
  // char const
  if Look = '\'' then
    nows = 1
    GetChar()
    if Look = '\\' then
      GetChar()
      if Look = 'n' then
        yytext[0] = 10
      else if Look = 't' then
        yytext[0] = 9
      else if Look = '\'' then
        yytext[0] = 39
      else if Look = '\\' then
        yytext[0] = 92
      else
        error("unknown \\ char")
      endif; endif; endif; endif
      GetChar()
      if Look = '\'' then
        GetChar()
        token[0] = yytext[0]
        return 'c'
      else
        error("missing \'")
      endif        
    else
      yytext[0] = Look
      GetChar()
      if Look = '\'' then
        GetChar()
        token[0] = yytext[0]
        return 'c'
      else
        error("missing \'")
      endif        
    endif
  endif   

  // array char const
  if Look = '"' then
    nows = 1; i = 0
    yytext[i] = Look
    i = i + 1
    GetChar()
    while Look # '"' do
      yytext[i] = Look
      i = i + 1
      GetChar()
      if i > 120 then
        error("string to long")
      endif
    wend
    yytext[i] = Look
    i = i + 1
    GetChar()
    yytext[i] = 0
    strcpy(token, yytext)
    return 's'
  endif
  
  // int const
  if isdigit(Look) then
    nows = 1; i = 0
    yytext[i] = Look
    i = i + 1
    GetChar()
    while isdigit(Look) do
      yytext[i] = Look
      i = i + 1
      GetChar()
      if i > 120 then
        error("number too long")
      endif
    wend
    yytext[i] = 0
    number = atoi(yytext)
    return 'n'
  endif
    
  // keyword or ident 
  if isalpha(Look) then
    nows = 1; i = 0
    yytext[i] = Look
    i = i + 1
    GetChar()
    while isalnum(Look) do
      yytext[i] = Look
      i = i + 1
      GetChar()
      if i > 120 then
        error("ident too long")
      endif
    wend
    yytext[i] = 0
    ch = gettoken(yytext)
    if ch = 'x' then
      strcpy(token, yytext)
    endif      
    return ch
  endif
  
  if Look < 0 then
    return Look     // EOF
  endif
  
  error("Unknown character")
end

// symbol table functions

// look up name in symbol table
// If found return symbol id else -1
func lookup(symtab: array char, name: array char): int
var cn: char; cs: char
  in: int; is: int
  found: int; sym: int
begin
  in = 0; cn = name[in]
  is = 0; cs = symtab[is]
  found = 1; sym = 0
  while cs # 0 do	// table end
    if cn # cs then
      found = 0
    endif
    is = is + 1; cs = symtab[is]
    if cn # 0 then
      in = in + 1; cn = name[in]
    endif
    if cs = ' ' then                    // symbol end
      if (found = 1) & (cn = 0) then
      	return sym
      else
	in = 0;  cn = name[in]
	is = is + 1; cs = symtab[is]
	sym = sym + 1
	found = 1
      endif
    endif
  wend
  return -1
end

func gettoken(name: array char): char
var id: int
begin  
  id = lookup(keywords, name)
  if id >= 0 then
    return tokens[id]
  else
    return 'x'
  endif
end

// offset is output variable
func gettype(offset: array int, name: array char): int
var id: int
begin
  id = lookup(locals, name)
  if id >= 0 then
    offset[0] = localoffset[id]
    return localtype[id]
  endif
  id = lookup(globals, name)
  if -1 = id then 
    error("variable not found")
  endif
  return globaltype[id]
end

func addsym(symtab: array char, name: array char): void
begin
  if strlen(symtab) + strlen(name) + 2 <= 1000 then
    strcat(symtab, name)
    strcat(symtab, " ")
  else
    error("symtab full")
  endif
end

func newname(symtab: array char, name: array char): int
var ident: int
begin
  ident = lookup(symtab, name)
  if ident >= 0 then
    return ident
  else
    addsym(symtab, name)
    return lookup(symtab, name)
  endif
end

func addglobal(name: array char, type: char): int
var id: int
begin
  id = newname(globals, name)
  globaltype[id] = type
  return id
end

func addlocal(name: array char, type: char, offset: int): int
var id: int
begin
  id = newname(locals, name)
  localtype[id] = type
  localoffset[id] = offset
  return id
end

func localsInit(): void
begin
  locals[0] = 0
end

func symtabInit(): void
begin
  strcpy(keywords, "array begin break char do end endif forward func if int else return then var void wend while ")
  strcpy(tokens, "abBCdeEfFiIlrtvVwW")
  globals[0] = 0
  localsInit()
end

func getname(symtab: array char, name: array char, ident: int): void  
var i: int; j: int
begin
  i = 0
  while ident > 0 do
    if ' ' = symtab[i] then
      ident = ident - 1
    endif
    i = i + 1
  wend
  j = 0
  while symtab[i] # ' ' do
    name[j] = symtab[i]
    j = j + 1
    i = i + 1
  wend
  name[j] = 0
end


func main(): int
var ch: int
begin  
  ch = 0
  scannerInit()
  symtabInit()
  while 1 do
    ch = yylex()
    if ch < 0 then break; endif
    printf("<%c,%s> ", ch, yytext)
  wend 
end

