%{
  /* basico04.y
   * Tiny BASIC no GOTO parser
   *
   * by Andre Adrian
   *
   * version 11jun2006
   */ 
  #include <stdio.h>
  #include "codegen.h"
  #include "scanner.h"
  
  int yylex(void);
  void yyerror(char *);
%}

%%

/* based on Pascal BNF */

program:  
   block4  
   | variableDeclaration ';' block4
   ;
   
block4:  
   funcDeclaration ';'
   ;
   
variableDeclaration:  
   'v' globalDecl  
   | variableDeclaration ';' globalDecl
   ;
   
globalDecl:
   declid ':' 'I'               { declint(); }                      
   | declid ':' 'a' 'n' 'I'     { declarrint();  }         
   | declid ':' 'C'             { declchar(); }                     
   | declid ':' 'a' 'n' 'C'     { declarrchar();  }
   ;

declid:
   'x'                          { declid(); }

funcDeclaration:
   func
   | funcDeclaration ';' func
   ;

func:
   'F' funcid '(' paramList ')' ':' returnType ';' blockOrForward
   ;

funcid:
   'x'                          { funcid(); }
   ;

paramList:
   paramDecl
   | paramList ',' paramDecl
   |  	 /* empty set */
   ;

paramDecl:
   declid ':' 'I'               { paramint(); }                   
   | declid ':' 'a' 'I'         { paramarrint(); }            
   | declid ':' 'C'             { paramchar(); }                  
   | declid ':' 'a' 'C'         { paramarrchar(); }           
   ;
      
returnType:
   'I'                          { printf("\t\t\t\t#func returns int\n"); }        
   | 'C'                        { printf("\t\t\t\t#func returns char\n"); }       
   | 'V'                        { printf("\t\t\t\t#func returns void\n"); }       
   ;
   
blockOrForward:  
   pblock3                       
   | 'f'                        { printf("\t\t\t\t#forward decl\n"); }            
   ;
   
pblock3:  
   pblock4  
   | pvariableDeclaration ';' pblock4
   ;
   
pblock4:  
   beginl stmtList 'e'         { funcepilogue(); }
   ;
   
beginl:
   'b'                          { funcprologue(); }
   ;
   
pvariableDeclaration:  
   'v' localDecl  
   | pvariableDeclaration ';' localDecl
   ;
   
localDecl:
   declid ':' 'I'               { localint(); }                      
   | declid ':' 'a' 'n' 'I'     { localarrint();  }         
   | declid ':' 'C'             { localchar(); }                     
   | declid ':' 'a' 'n' 'C'     { localarrchar();  }
   ;

stmtList:  
   statement ';'
   | stmtList statement ';'
   ;  

statement:
   variable '=' expr              { assign(); }
   | 'i' expr thenl stmtList elseOrEndif
   | whilel expr dol stmtList 'w' { wendjump(); }
   | callid '(' exprList ')'      { callproc(); }
   | 'r'                          { retrn(); }
   | 'r' expr                     { retrnvalue(); }
   | 'B'                          { breakjump(); }
   |   	 /* empty set */
   ;
   
variable:
   varid                     
   | varid '[' expr ']'      { assignarray(); }        
   ;  

varid:
   'x'                       { varid(); }             
    
whilel:
   'W'             	     { whilelabel(); }        
   ;

dol:
   'd'             	     { dojump(); }            
   ;

thenl:
   't'                       { thenjump(); }          

elseOrEndif:
   'E'                       { endiflabel(); }        
   | elsel stmtList 'E'     { endiflabel(); }        
   ;
   
elsel:
   'l'                       { elsejump(); }          
   ;

callid:
   'x'                       { callid(); }            
   ;
   
exprList:
   expr                      { callarg(); }                    
   | exprList ',' expr      { callarg(); }                    
   |  	 /* empty set */
   ;
  
expr:
   addexpr
   | expr '=' addexpr        { eq(); }                                     
   | expr '#' addexpr        { ne(); }                                     
   | expr '<' addexpr        { lt(); }                                     
   | expr '<' '=' addexpr    { le(); }    
   | expr '>' addexpr        { gt(); }                                     
   | expr '>' '=' addexpr    { ge(); }    
   ;

addexpr:
   term
   | addexpr '+' term        { add(); }     
   | addexpr '-' term        { sub(); }    
   | addexpr '|' term        { or(); }     
   | addexpr '^' term        { xor(); }     
   ;

term:
   unaryFact
   | term '*' unaryFact     { mul(); }                      
   | term '/' unaryFact     { div(); }                      
   | term '%' unaryFact     { mod(); }                      
   | term '&' unaryFact     { and(); }                      
   ;

unaryFact: 
   '-' fact                 { neg(); }                          
   | '~' fact               { not(); }                          
   | fact
   ;

fact:
   'n'      	      	      { pushconst(); }                     
   | 'c'      	              { pushchrconst(); }                     
   | 's'      	              { pushstrconst(); }                     
   | 'x'                      { pushvar(); }                       
   | arrayid '[' expr ']'     { pusharrayvar(); }            
   | '(' expr ')'
   | callid '(' exprList ')'  { callfunc(); }             
   ;
   
arrayid:
   'x'                        { arrayid(); }
   ;   

%%

void yyerror(char *s) {
    fprintf(stderr, "%s\n", s);
}

int main(void) {
    scannerInit();
    symtabInit();
    fileprologue();
#if YYDEBUG
    yydebug = 1;
#endif    
    yyparse();
    fileepilogue();
    return 0;
}
