/* codegen.c
 * GNU as / Intel 386 32bit mode Code Generator 
 * Input: Stack-machine Opcodes
 * Output: Opcodes for Registers as Stack cache with Register renaming
 *
 * by Andre Adrian
 *
 * version 19jun2006
 *
 * compiler bug: srand(time(0)) not working
 * nested funcs need stacked oldse?
 */

#include <stdio.h>
#include "scanner.h"

// uncomment to test stand-alone
// #define MODULTEST

#define REGS 4          // cx, bx, ax
#define LEVELS 10       // if, while nesting level

extern char token[128]; // import from basico03.l or scanner.c
extern int number;      // import from basico03.l or scanner.c

int se;                 // Stack elements
int oldse;
char reg[REGS+1];         // Register usage

int label;    	      	// Label-id generator
int L1[LEVELS], L2[LEVELS]; // if jump label stack
int L3[LEVELS], L4[LEVELS]; // while jump label stack
int level;              // for if
int whilelevel;         // for while 
int lco;                // String const id
char iddecl[32];        // variable definition under construction
char idfunc[32];        // function definition under construction
char idcall[32];        // function call under construction
char idvar[32];         // assignment under construction
char idarray[32];       // variable access under construction
int args;      
int isarray;
int frame;              // locals
int param;        
int offsetvar[1];       // offset of assignment
char typevar;           // type of assignment

void cacheinit() {
  se = 0;
  strcpy(reg, "dcba");
}

void fileprologue() {
  cacheinit();
  token[127] = '\0';
  label = 0;
  level = -1;
  whilelevel = -1;
  lco = 0;
  printf("\t.file  \"a.bas\"\n");
  printf("\t.text\n");
}

void fileepilogue() {
  printf("\t.section .note.GNU-stack,\"\",@progbits\n");
  printf("\t.ident \"Basico 0.4\"\n");
}

void declid() {
  strcpy(iddecl, token);
}  

// type symbols:
// i=global integer, I=global integer array
// j=local integer, J=local integer array
// K=call-by-reference integer array
// c=global char, C=global char array
// d=local char, D=local char array
// E=call-by-reference char array

void declint() {
  printf("\t.comm  %s,4,4\n", iddecl);
  addglobal(iddecl, 'i');
}

void declarrint() {
  printf("\t.comm  %s,%d,4\n", iddecl, number*4);
  addglobal(iddecl, 'I');
}

void declchar() {
  printf("\t.comm  %s,1,1\n", iddecl);
  addglobal(iddecl, 'c');
}

void declarrchar() {
  printf("\t.comm  %s,%d,1\n", iddecl, number);
  addglobal(iddecl, 'C');
}

void localint() {
  frame = frame -4;
  addlocal(iddecl, 'j', frame);
}

void localarrint() {
  frame = frame - number*4;
  addlocal(iddecl, 'J', frame);
}

void localchar() {
  frame = frame - 1;
  addlocal(iddecl, 'd', frame);
}

void localarrchar() {
  frame = frame - number;
  addlocal(iddecl, 'D', frame);
}

void paramint() {
  addlocal(iddecl, 'j', param);
  param = param +4;
}

void paramarrint() {
  addlocal(iddecl, 'K', param);
  param = param +4;
}

void paramchar() {
  addlocal(iddecl, 'd', param);
  param = param +4;
}

void paramarrchar() {
  addlocal(iddecl, 'E', param);
  param = param +4;
}

void funcid() {
  strcpy(idfunc, token);
  frame = 0;
  param = 8;
}

void funcprologue() {
  cacheinit();
  printf(".globl\t%s\n", idfunc);
  printf("\t.type\t%s, @function\n", idfunc);
  printf("%s:\n", idfunc);
  printf("\tpushl\t%%ebp\n");
  printf("\tmovl\t%%esp, %%ebp\n");
  if (frame < 0) {
    printf("\tsubl\t$%d, %%esp\n", -frame);
  }
}

void funcepilogue() {
  printf("\tleave\n");
  printf("\tret\n");
  printf("\t.size %s, .-%s\n", idfunc, idfunc);
  localsInit();
}

void retrn() {
  printf("\tleave\n");
  printf("\tret\n");
}

void retrnvalue() {
  // return value in reg eax
  if (reg[0] != 'a') {
    printf("\tmovl   %%e%cx, %%eax\n", reg[0]);
  }
  retrn();
}

pushcache() {
  int i;
  
  i = se;
  while (i >= 1) {
    i = i - 1;
    printf("\tpushl\t%%e%cx\n", reg[i]);
  }
}

popcache() {
  int i;
  
  i = 1;
  while (i < se) {
    printf("\tpopl\t%%e%cx\n", reg[i]);
    i = i + 1;
  }
}

void callid() {
  pushcache();
  oldse = se;
  se = 0;
  // printf("\tsubl\t$12, %%esp\n");
  strcpy(idcall, token);
  args = 0;
}
  
void callarg() {
  ++args;
  if (args > REGS) {
    error("too many args");
  }
}

void call() {
  // C pushes args from right to left
  int i;
  
  i = 0;
  while (i < se) {
    printf("\tpushl\t%%e%cx\n", reg[i]);
    i = i + 1;
  }
  printf("\tcall\t%s\n", idcall);
  if (args > 0) {
    printf("\taddl\t$%d, %%esp\n", 4*args);
  }
}

void callproc() {
  call();
  cacheinit();
}

void callfunc() {
  call();
  se = oldse+1;
  strcpy(reg, "abcd");  // return value in eax
  popcache();
}

void thenjump() {
  // L1, L2 are local variables in a recursive decent parser
  ++level;
  L1[level] = label++;
  L2[level] = L1[level];
  printf("\tandl\t%%e%cx, %%e%cx", reg[0], reg[0]);
  printf("\t#then_jump %d\n", level);
  printf("\tjz\t.L%d\n", L1[level]);
  cacheinit();
}

void elsejump() {
  L2[level] = label++;
  printf("\tjmp\t.L%d", L2[level]);
  printf("\t\t#else_jump %d\n", level); 
  printf(".L%d:\n", L1[level]);
}

void endiflabel() {
  printf(".L%d:", L2[level]);
  printf("\t\t\t\t#endif_label %d\n", level);
  --level;
}

void whilelabel() {
  // L3, L4 are local variables in a recursive decent parser
  ++whilelevel;
  L3[whilelevel] = label++;
  L4[whilelevel] = label++;
  printf(".L%d:", L3[whilelevel]);
  printf("\t\t\t\t#while_label %d\n", whilelevel);
}

void dojump() {
  // L3, L4 are local variables in a recursive decent parser
  printf("\tandl\t%%e%cx, %%e%cx", reg[0], reg[0]);
  printf("\t#do_jump %d\n", whilelevel);
  printf("\tjz\t.L%d\n", L4[whilelevel]);
  cacheinit();
}

void breakjump() {
  if (whilelevel < 0) {
    error("break without while");
  }
  printf("\tjmp\t.L%d", L4[whilelevel]);
  printf("\t#break_jump %d\n", whilelevel);
}
  
void wendjump() {
  printf("\tjmp\t.L%d", L3[whilelevel]);
  printf("\t\t#wend_jump %d\n", whilelevel); 
  printf(".L%d:\n", L4[whilelevel]);
  --whilelevel;
}

void push() {
  char tmp;
  int i;
  i = REGS-1;
  if (se >= REGS) {
    printf("\tpushl\t%%e%cx\n", reg[i]);
  }
  tmp = reg[i];
  while (i > 0) {
    reg[i] = reg[i - 1];
    i = i - 1;
  }
  reg[0] = tmp; 
  ++se;
}

void pop() {
  char tmp;
  int i;
  --se;
  if (se == 0) {
    cacheinit();
  } else {
    i = 0;
    tmp = reg[i];
    while (i < REGS-1) {
      reg[i] = reg[i + 1];
      i = i + 1;
    }
    reg[i] = tmp; 
  }
  if (se >= REGS) {
    printf("\tpopl\t%%e%cx\n", reg[i]);
  }
}

void arrayid() {
  strcpy(idarray, token);
}  

varid() {
  isarray = 0;
  strcpy(idvar, token);
  typevar = gettype(offsetvar, idvar);
}

assignarray() {
  isarray = 1;
}

assign() {
  if ('i' == typevar) {
    printf("\tmovl\t%%e%cx, %s\n", reg[0], idvar);
  } else if ('j' == typevar) {
    printf("\tmovl\t%%e%cx, %d(%%ebp)\n", reg[0], offsetvar[0]);
  } else if ('I' == typevar) {
    printf("\tmovl\t%%e%cx, %s(,%%e%cx,4)\n", reg[0], idvar, reg[1]);
    pop();    
  } else if ('J' == typevar) {
    printf("\tmovl\t%%e%cx, %d(%%ebp,%%e%cx,4)\n", reg[0], offsetvar[0], reg[1]);
    pop();    
  } else if ('K' == typevar) {
    printf("\tmovl\t%d(%%ebp), %%esi\n", offsetvar[0]);
    printf("\tmovl\t%%e%cx, (%%esi,%%e%cx,4)\n", reg[0], reg[1]);
    pop();    
  } else if ('c' == typevar) {
    printf("\tmovb\t%%%cl, %s\n", reg[0], idvar);
  } else if ('d' == typevar) {
    printf("\tmovb\t%%%cl, %d(%%ebp)\n", reg[0], offsetvar[0]);
  } else if ('C' == typevar) {
    printf("\tmovb\t%%%cl, %s(%%e%cx)\n", reg[0], idvar, reg[1]);
    pop();    
  } else if ('D' == typevar) {
    printf("\tmovb\t%%%cl, %d(%%e%cx,%%ebp)\n", reg[0], offsetvar[0], reg[1]);
    pop();    
  } else if ('E' == typevar) {
    printf("\tmovl\t%d(%%ebp), %%esi\n", offsetvar[0]);
    printf("\tmovb\t%%%cl, (%%esi,%%e%cx)\n", reg[0], reg[1]);
    pop();    
  }  
  pop();
}

pushconst() {
  push();
  printf("\tmovl\t$%d, %%e%cx\n", number, reg[0]);
}

pushchrconst() {
  push();
  printf("\tmovl\t$%d, %%e%cx", token[0], reg[0]);
  if (token[0] >= ' ') {
    printf("\t# '%c'\n", token[0]);
  } else {
    printf("\t# chr(%d)\n", token[0]);
  }  
}

pushstrconst() {
  printf("\t.section .rodata\n");
  printf(".LC%d:\n", lco);
  printf("\t.string\t%s\n", token);
  printf("\t.text\n");
  push();
  printf("\tmovl\t$.LC%d, %%e%cx\n", lco, reg[0]);
  ++lco;
}

pushvar() {
  int type;
  int offset[1];          // array for call-by-reference
  
  type = gettype(offset, token);
  push();
  if ('i' == type) {
    printf("\tmovl\t%s, %%e%cx\n", token, reg[0]);
  } else if ('I' == type || 'C' == type) {
    printf("\tmovl\t$%s, %%e%cx\n", token, reg[0]);
  } else if ('j' == type) {
    printf("\tmovl\t%d(%%ebp), %%e%cx\n", offset[0], reg[0]);
  } else if ('J' == type || 'D' == type) {
    printf("\tleal\t%d(%%ebp), %%e%cx\n", offset[0], reg[0]);
  } else if ('c' == type) {
    printf("\tmovzbl\t%s, %%e%cx\n", token, reg[0]);
  } else if ('d' == type) {
    printf("\tmovzbl\t%d(%%ebp), %%e%cx\n", offset[0], reg[0]);
  } else if ('K' == type || 'E' == type) {
    printf("\tmovl\t%d(%%ebp), %%e%cx\n", offset[0], reg[0]);
  }  
}

pusharrayvar() {
  int type;
  int offset[1];          // array for call-by-reference
  
  type = gettype(offset, idarray);
  if ('I' == type) {
    printf("\tmovl\t%s(,%%e%cx,4), %%e%cx\n", idarray, reg[0], reg[0]);
  } else if ('J' == type) {
    printf("\tmovl\t%d(%%ebp,%%e%cx,4), %%e%cx\n", offset[0], reg[0], reg[0]);
  } else if ('K' == type) {
    printf("\tmovl\t%d(%%ebp), %%esi\n", offset[0]);
    printf("\tmovl\t(%%esi,%%e%cx,4), %%e%cx\n", reg[0], reg[0]);    
  } else if ('C' == type) {
    printf("\tmovzbl\t%s(%%e%cx), %%e%cx\n", idarray, reg[0], reg[0]);
  } else if ('D' == type) {
    printf("\tmovzbl\t%d(%%e%cx,%%ebp), %%e%cx\n", offset[0], reg[0], reg[0]);
  } else if ('E' == type) {
    printf("\tmovl\t%d(%%ebp), %%esi\n", offset[0]);
    printf("\tmovzbl\t(%%esi,%%e%cx), %%e%cx\n", reg[0], reg[0]);    
  }  
}

add() {
  printf("\taddl\t%%e%cx, %%e%cx\n", reg[0], reg[1]);
  pop();
}

sub() {
  printf("\tsubl\t%%e%cx, %%e%cx\n", reg[0], reg[1]);
  pop();
}

mul() {
  printf("\timul\t%%e%cx, %%e%cx\n", reg[0], reg[1]);
  pop();
}  

divmod() {
  // idiv op-code uses edx and eax for 64bit dividend at input
  // output is quotient in eax, remainder in edx
  int tmp;
  int i;
  
  i = 0;
  while (i < REGS) {
    if ((i != 1) && ('a' == reg[i])) {
      printf("\txchgl\t%%e%cx, %%e%cx\n", reg[i], reg[1]);
      tmp = reg[i]; 
      reg[i] = reg[1];
      reg[1] = tmp;
      break;
    }
    i = i + 1;
  }
  printf("\tmovl\t%%edx, %%esi\n");   // save edx
  printf("\tcltd\n");
  printf("\tidivl\t%%e%cx\n", reg[0]);
} 
 
div() {
  divmod();
  printf("\tmovl\t%%esi, %%edx\n");   // restore edx
  pop();
} 
 
mod() {
  divmod();
  printf("\tmovl\t%%edx, %%eax\n");   // use remainder
  printf("\tmovl\t%%esi, %%edx\n");   // restore edx
  pop();
}

neg() {
  printf("\tnegl\t%%e%cx\n", reg[0]);
}

and() {
  printf("\tandl\t%%e%cx, %%e%cx\n", reg[0], reg[1]);
  pop();
}

or() {
  printf("\torl\t%%e%cx, %%e%cx\n", reg[0], reg[1]);
  pop();
}

xor() {
  printf("\txorl\t%%e%cx, %%e%cx\n", reg[0], reg[1]);
  pop();
}

not() {
  printf("\tnotl\t%%e%cx\n", reg[0]);
}

eq() {
  printf("\tcmpl\t%%e%cx, %%e%cx", reg[0], reg[1]);
  printf("\t#equal\n");
  printf("\tsete\t%%%cl\n", reg[1]);
  printf("\tmovzbl\t%%%cl, %%e%cx\n",reg[1], reg[1]);
  pop();
}  
  
ne() {
  printf("\tcmpl\t%%e%cx, %%e%cx", reg[0], reg[1]);
  printf("\t#not equal\n");
  printf("\tsetne\t%%%cl\n", reg[1]);
  printf("\tmovzbl\t%%%cl, %%e%cx\n",reg[1], reg[1]);
  pop();
}  
  
lt() {
  printf("\tcmpl\t%%e%cx, %%e%cx", reg[0], reg[1]);
  printf("\t#less\n");
  printf("\tsetl\t%%%cl\n", reg[1]);
  printf("\tmovzbl\t%%%cl, %%e%cx\n",reg[1], reg[1]);
  pop();
}  
  
le() {
  printf("\tcmpl\t%%e%cx, %%e%cx", reg[0], reg[1]);
  printf("\t#less or equal\n");
  printf("\tsetle\t%%%cl\n", reg[1]);
  printf("\tmovzbl\t%%%cl, %%e%cx\n",reg[1], reg[1]);
  pop();
}  
  
gt() {
  printf("\tcmpl\t%%e%cx, %%e%cx", reg[0], reg[1]);
  printf("\t#greater\n");
  printf("\tsetg\t%%%cl\n", reg[1]);
  printf("\tmovzbl\t%%%cl, %%e%cx\n",reg[1], reg[1]);
  pop();
}  
  
ge() {
  printf("\tcmpl\t%%e%cx, %%e%cx", reg[0], reg[1]);
  printf("\t#greater or equal\n");
  printf("\tsetge\t%%%cl\n", reg[1]);
  printf("\tmovzbl\t%%%cl, %%e%cx\n",reg[1], reg[1]);
  pop();
}  
  

#ifdef MODULTEST
main() {
  // c = a - b
  pushvar('a');
  pushvar('b');
  sub();
  pop('c');
  
  pushvar('a');
  pushvar('b');
  pushvar('c');
  pushvar('d');
  pushvar('e');
  pushvar('f');
  sub();
  pop('x');
  sub();
  pop('y');
  sub();
  pop('z');
  
  // a = a - 1
  pushvar('a');
  pushvar('1');
  sub();
  pop('a');
}  
#endif   
